BeetLeaves = {}

local modDirectory = g_currentModDirectory

local function registerFillTypes()
    local hudFile = "textures/hud_fill_beetLeaves.png"
    local hudFileSmall = "textures/hud_fill_beetLeaves_sml.png"

    local price = g_fillTypeManager:getFillTypeByName("GRASS_WINDROW").pricePerLiter
    local mass = g_fillTypeManager:getFillTypeByName("GRASS_WINDROW").massPerLiter

    g_fillTypeManager:addFillType("BEETLEAVES", g_i18n:getText("fillType_beetLeaves"), true, price, mass, 50, hudFile, hudFileSmall, modDirectory, nil, { 1, 1, 1 }, nil, false)
	--g_fillTypeManager:addFillType(name, title, showOnPriceTable, pricePerLiter, massPerLiter, maxPhysicalSurfaceAngle, hudOverlayFilename, hudOverlayFilenameSmall, modDirectory, nil, {1,1,1}, palletFilename, false);
end

local function initFillTypes(mission, node)
	local fillTypeIndex = g_fillTypeManager:getFillTypeIndexByName("BEETLEAVES")
	
	g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["BULK"])
	g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["SILAGETRAILER"])
	g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["FORK"])

	g_fillTypeManager.fillTypeToSample[fillTypeIndex] = g_fillTypeManager.fillTypeToSample[g_fillTypeManager:getFillTypeIndexByName("CHAFF")]	
	
	local diffuseMapFilename = Utils.getFilename("textures/beetLeaves_diffuse.png", modDirectory)
    local normalMapFilename = Utils.getFilename("textures/beetLeaves_normal.png", modDirectory)
    local distanceMapFilename = Utils.getFilename("textures/beetLeaves_distance_diffuse.png", modDirectory)

	g_densityMapHeightManager:addDensityMapHeightType("BEETLEAVES", math.rad(35), 6.0, 0.35, 0.10, 0.10, 1.20, false, diffuseMapFilename, normalMapFilename, distanceMapFilename, false)
	--g_densityMapHeightManager:addDensityMapHeightType(fillTypeName, maxSurfaceAngle, collisionScale, collisionBaseOffset, minCollisionOffset, maxCollisionOffset, fillToGroundScale, allowsSmoothing, diffuseMapFilename, normalMapFilename, distanceFilename, isBaseType)

	g_currentMission.hud.fillLevelsDisplay:refreshFillTypes(g_fillTypeManager)
	
    local beetLeaves_materialHolderFilename = Utils.getFilename("materialHolders/beetLeaves_materialHolder.i3d", modDirectory)
    BeetLeaves.materialHolderFilename = loadI3DFile(beetLeaves_materialHolderFilename, false, true, false)
	
	if g_modIsLoaded["FS19_Pitchfork"] then
		local pitchfork = _G["FS19_Pitchfork"].Pitchfork
		local function addBeetLeavesSupport(self, xmlFile, configId)
			if self:getFillUnitSupportsFillType(self.fillableToolFillUnitIndex, FillType.CHAFF) then
				local spec = self.spec_fillUnit
				spec.fillUnits[self.fillableToolFillUnitIndex].supportedFillTypes[FillType.BEETLEAVES] = true
			end
		end
		pitchfork.loadFillableHandToolConfiguration = Utils.appendedFunction(pitchfork.loadFillableHandToolConfiguration, addBeetLeavesSupport)
	end
end

local function unload()
	if BeetLeaves.materialHolderFilename ~= nil then
		delete(BeetLeaves.materialHolderFilename)
		BeetLeaves.materialHolderFilename = nil
	end
end

local function sellingStationLoad(self, xmlFile, key, customEnvironment)
	if self ~= nil then
		if self.acceptedFillTypes[g_fillTypeManager:getFillTypeIndexByName("GRASS_WINDROW")] then
			local beetLeavesIndex = g_fillTypeManager:getFillTypeIndexByName("BEETLEAVES")
			
			self:addAcceptedFillType(beetLeavesIndex, g_fillTypeManager:getFillTypeByName("BEETLEAVES").pricePerLiter, true, false)
			self.priceMultipliers[beetLeavesIndex] = self.priceMultipliers[g_fillTypeManager:getFillTypeIndexByName("GRASS_WINDROW")]
			
			self:initPricingDynamics()
		end
	end
end

local function bgaLoad(self, superFunc, xmlFile, key, customEnvironment)
    local returnValue = superFunc(self, xmlFile, key, customEnvironment)
	
	if self ~= nil then
		local grassIndex = g_fillTypeManager:getFillTypeIndexByName("GRASS_WINDROW")
		local beetLeavesIndex = g_fillTypeManager:getFillTypeIndexByName("BEETLEAVES")
		
		for _, slot in pairs(self.bunker.slots) do
			if slot.fillTypes[grassIndex] then
				local fillType = {}
				fillType.fillLevel = 0
				fillType.scale = slot.fillTypes[grassIndex].scale
				fillType.pricePerLiter = slot.fillTypes[grassIndex].pricePerLiter
				
				slot.fillTypes[beetLeavesIndex] = fillType

				if self.bunker.fillTypeToSlot[beetLeavesIndex] == nil then
					self.bunker.fillTypeToSlot[beetLeavesIndex] = slot
				end
			end
		end
	end
	
	return returnValue
end

local function bunkerSiloLoad(self, superFunc, xmlFile, key, customEnvironment)
	local returnValue = superFunc(self, xmlFile, key, customEnvironment)
	
	if self.acceptedFillTypes[g_fillTypeManager:getFillTypeIndexByName("CHAFF")] then
		self.acceptedFillTypes[g_fillTypeManager:getFillTypeIndexByName("BEETLEAVES")] = true
	end
	
	return returnValue
end

local function init()
    FillTypeManager.loadMapData = Utils.appendedFunction(FillTypeManager.loadMapData, registerFillTypes)
	FSBaseMission.loadMap = Utils.prependedFunction(FSBaseMission.loadMap, initFillTypes)
	FSBaseMission.delete = Utils.appendedFunction(FSBaseMission.delete, unload)
	
	SellingStation.load = Utils.appendedFunction(SellingStation.load, sellingStationLoad)
	Bga.load = Utils.overwrittenFunction(Bga.load, bgaLoad)
	
	BunkerSilo.load = Utils.overwrittenFunction(BunkerSilo.load, bunkerSiloLoad)
end

init()