--	Agromet-Famarol Z-413 "Neptun"
--
--	Author: Rockstar94
--	Date: 21.04.2020
--

Neptun = {}
Neptun.modName = g_currentModName
Neptun.modDirectory = g_currentModDirectory

source(Neptun.modDirectory .. "scripts/events/SetSensorStateEvent.lua")
source(Neptun.modDirectory .. "scripts/events/SetUnfoldStateEvent.lua")
source(Neptun.modDirectory .. "scripts/events/SetUnloadStateEvent.lua")

function Neptun.prerequisitesPresent(specializations)
	return SpecializationUtil.hasSpecialization(FruitPreparer, specializations) and  SpecializationUtil.hasSpecialization(Combine, specializations)
end

function Neptun.registerOverwrittenFunctions(vehicleType)
	SpecializationUtil.registerOverwrittenFunction(vehicleType, "processFruitPreparerArea", Neptun.processFruitPreparerArea)
	SpecializationUtil.registerOverwrittenFunction(vehicleType, "getCanBeTurnedOn", Neptun.getCanBeTurnedOn)
	SpecializationUtil.registerOverwrittenFunction(vehicleType, "getTurnedOnNotAllowedWarning", Neptun.getTurnedOnNotAllowedWarning)
	SpecializationUtil.registerOverwrittenFunction(vehicleType, "getCanToggleDischargeToObject", Neptun.getCanToggleDischargeToObject)
	SpecializationUtil.registerOverwrittenFunction(vehicleType, "handleDischargeRaycast", Neptun.handleDischargeRaycast)
end

function Neptun.registerEventListeners(vehicleType)
	SpecializationUtil.registerEventListener(vehicleType, "onLoad", Neptun)
	SpecializationUtil.registerEventListener(vehicleType, "onPostLoad", Neptun)
	SpecializationUtil.registerEventListener(vehicleType, "onDelete", Neptun)
	SpecializationUtil.registerEventListener(vehicleType, "onUpdateTick", Neptun)
	SpecializationUtil.registerEventListener(vehicleType, "onTurnedOn", Neptun)
	SpecializationUtil.registerEventListener(vehicleType, "onTurnedOff", Neptun)
	SpecializationUtil.registerEventListener(vehicleType, "onReadStream", Neptun)
	SpecializationUtil.registerEventListener(vehicleType, "onWriteStream", Neptun)
	SpecializationUtil.registerEventListener(vehicleType, "onReadUpdateStream", Neptun)
	SpecializationUtil.registerEventListener(vehicleType, "onWriteUpdateStream", Neptun)
	SpecializationUtil.registerEventListener(vehicleType, "onRegisterActionEvents", Neptun)
	SpecializationUtil.registerEventListener(vehicleType, "saveToXMLFile", Neptun)
	SpecializationUtil.registerEventListener(vehicleType, "onAIImplementStart", Neptun)
	SpecializationUtil.registerEventListener(vehicleType, "onAIImplementEnd", Neptun)
end

function Neptun.registerFunctions(vehicleType)
	SpecializationUtil.registerFunction(vehicleType, "setLeavesTankIsFilling", Neptun.setLeavesTankIsFilling)
	SpecializationUtil.registerFunction(vehicleType, "loadFillEffects", Neptun.loadFillEffects)
	SpecializationUtil.registerFunction(vehicleType, "loadTerrainScanner", Neptun.loadTerrainScanner)
	SpecializationUtil.registerFunction(vehicleType, "loadDischargeParameters", Neptun.loadDischargeParameters)
	SpecializationUtil.registerFunction(vehicleType, "getCanSensorPickupState", Neptun.getCanSensorPickupState)
	SpecializationUtil.registerFunction(vehicleType, "setSensorState", Neptun.setSensorState)
	SpecializationUtil.registerFunction(vehicleType, "setUnfoldState", Neptun.setUnfoldState)
	SpecializationUtil.registerFunction(vehicleType, "dischargeLeaves", Neptun.dischargeLeaves)
	SpecializationUtil.registerFunction(vehicleType, "setDischargeLeavesState", Neptun.setDischargeLeavesState)
	SpecializationUtil.registerFunction(vehicleType, "updateDischargeLeavesSound", Neptun.updateDischargeLeavesSound)
end

function Neptun:getCanBeTurnedOn(superFunc)
	local spec = self.spec_fruitPreparer
	if spec.turnOffIfFull then
		if self:getFillUnitFreeCapacity(spec.fillUnitIndex) == 0 then
			if self:getIsAIActive() then
				local rootVehicle = self:getRootVehicle()
				rootVehicle:stopAIVehicle(AIVehicle.STOP_REASON_GRAINTANK_IS_FULL)
			end
			return false
		end
	end
	return superFunc(self)
end

function Neptun:getTurnedOnNotAllowedWarning(superFunc)
	local spec = self.spec_fruitPreparer
	if spec.turnOffIfFull then
		if self:getFillUnitFreeCapacity(spec.fillUnitIndex) == 0 then
			return spec.unloadWarning
		end
	end
	return superFunc(self)
end

function Neptun:onAIImplementStart()
	local spec = self.spec_fruitPreparer
	if spec.restoreLastState then
		spec.lastUnloadingState = spec.isManualUnloding
		spec.lastSensorState = spec.isSensorLowered
	end
	if spec.isManualUnloding then
		self:setUnfoldState(false)
	end
end

function Neptun:onAIImplementEnd()
	local spec = self.spec_fruitPreparer
	if spec.restoreLastState then
		self:setUnfoldState(spec.lastUnloadingState)
		self:setSensorState(spec.lastSensorState)
	end
end

function Neptun:onRegisterActionEvents(isActiveForInput, isActiveForInputIgnoreSelection)
	if self.isClient then
		local spec = self.spec_fruitPreparer
        self:clearActionEventsTable(spec.actionEvents)
        if isActiveForInputIgnoreSelection then
            local _, actionEventId = self:addActionEvent(spec.actionEvents, InputAction.NEPTUN_SENSOR, self, Neptun.actionEventToggleFoldState, false, true, false, true, nil)
            g_inputBinding:setActionEventTextPriority(actionEventId, GS_PRIO_HIGH)
			Neptun.updateFoldStateActionEvent(self)
			
            local _, actionEventId2 = self:addActionEvent(spec.actionEvents, InputAction.NEPTUN_UNLOAD_MODE, self, Neptun.actionEventUnloadMode, false, true, false, true, nil)
            g_inputBinding:setActionEventTextPriority(actionEventId2, GS_PRIO_HIGH)
			Neptun.updateUnloadModeActionEvent(self)
			
            local _, actionEventId3 = self:addActionEvent(spec.actionEvents, InputAction.NEPTUN_UNLOAD, self, Neptun.actionEventUnloadLeaves, false, true, false, true, nil)
            g_inputBinding:setActionEventTextPriority(actionEventId3, GS_PRIO_HIGH)
			g_inputBinding:setActionEventText(actionEventId3, g_i18n:getText("neptun_leavesUnload"))
        end
    end
end

function Neptun.actionEventUnloadLeaves(self, actionName, inputValue, callbackState, isAnalog)
	self:setDischargeLeavesState(true)
end

function Neptun.actionEventUnloadMode(self, actionName, inputValue, callbackState, isAnalog)
	local spec = self.spec_fruitPreparer
	self:setUnfoldState(not spec.isManualUnloding)
end

function Neptun.actionEventToggleFoldState(self, actionName, inputValue, callbackState, isAnalog)
    local spec = self.spec_fruitPreparer
    if self:getCanSensorPickupState() then
        self:setSensorState(not spec.isSensorLowered)
    end
end

function Neptun.updateUnloadLeavesActionEvent(self)
    local spec = self.spec_fruitPreparer
	local actionEvent = spec.actionEvents[InputAction.NEPTUN_UNLOAD]
	
	local eventActive = false
	local dischargeNode = spec.discharge
	if dischargeNode ~= nil then
		local attacherVehicle = self:getAttacherVehicle()
		local allowUnload = attacherVehicle ~= nil and attacherVehicle:getIsMotorStarted() ~= nil and attacherVehicle:getIsMotorStarted()
		if allowUnload then
			local allowedToDischarge = self:getCanDischargeToLand(dischargeNode) and self:getCanDischargeAtPosition(dischargeNode)
			local canDischargeToGround = self:getCanDischargeToGround(dischargeNode)
			local foldState = not spec.isSensorLowered and spec.isManualUnloding and not spec.unloadState
			if allowedToDischarge and canDischargeToGround and foldState then
				local fillLevelPercentage = self:getFillUnitFillLevelPercentage(spec.fillUnitIndex)
				if fillLevelPercentage > spec.discharge.manualDischargeMinFillLevel then
					eventActive = true
				end
			end
		end
	end
	
	if actionEvent ~= nil then	
		g_inputBinding:setActionEventActive(actionEvent.actionEventId, eventActive)
		g_inputBinding:setActionEventTextVisibility(actionEvent.actionEventId, eventActive)
	end
end

function Neptun.updateUnloadModeActionEvent(self)
    local spec = self.spec_fruitPreparer
	local actionEvent = spec.actionEvents[InputAction.NEPTUN_UNLOAD_MODE]
	if actionEvent ~= nil then
		if spec.isManualUnloding then
			g_inputBinding:setActionEventText(actionEvent.actionEventId, string.format(g_i18n:getText("neptun_leavesUnloadOption"), g_i18n:getText("neptun_leavesUnloadOption_manual")))
		else
			g_inputBinding:setActionEventText(actionEvent.actionEventId, string.format(g_i18n:getText("neptun_leavesUnloadOption"), g_i18n:getText("neptun_leavesUnloadOption_automatic")))
		end
	end
end

function Neptun.updateFoldStateActionEvent(self)
    local spec = self.spec_fruitPreparer
    if spec.animationName ~= "" then
        local actionEvent = spec.actionEvents[InputAction.NEPTUN_SENSOR]
        if actionEvent ~= nil then
            if spec.isSensorLowered then
                g_inputBinding:setActionEventText(actionEvent.actionEventId, g_i18n:getText("neptun_leavesSensor_fold"))
            else
                g_inputBinding:setActionEventText(actionEvent.actionEventId, g_i18n:getText("neptun_leavesSensor_unfold"))
            end
            g_inputBinding:setActionEventActive(actionEvent.actionEventId, self:getCanSensorPickupState())
			g_inputBinding:setActionEventTextVisibility(actionEvent.actionEventId, self:getCanSensorPickupState())
        end
    end
end

function Neptun:onLoad(savegame)
	local spec = self.spec_fruitPreparer
	local key = "vehicle.neptun"
	
	spec.actionEvents = {}
	
	spec.turnOffIfFull = Utils.getNoNil(getXMLBool(self.xmlFile, key .. "#turnOffIfFull"), false)
	spec.fillUnitIndex = Utils.getNoNil(getXMLInt(self.xmlFile, key .. "#fillUnitIndex"), 2)
	spec.unloadInfoIndex = Utils.getNoNil(getXMLInt(self.xmlFile, key .. "#unloadInfoIndex"), 1)
	
	local fillTypeName = Utils.getNoNil(getXMLString(self.xmlFile, key .. "#leavesFillType"), "CHAFF")
	local fillType = g_fillTypeManager:getFillTypeByName(fillTypeName)
	spec.leavesFillType = fillType.index

	spec.unloadWarning = g_i18n:getText("neptun_unloadWarning")

	spec.fillMultiplier = Utils.getNoNil(getXMLFloat(self.xmlFile, key .. "#fillMultiplier"), 1)

	spec.terrainScanner = {}
	self:loadTerrainScanner(self.xmlFile, key .. ".terrainScanner", spec.terrainScanner)
	
	spec.discharge = {}
	self:loadDischargeParameters(self.xmlFile, key .. ".dischargeLeaves", spec.discharge)
	
	self:loadFillEffects(self.xmlFile, key, spec)
	
    spec.lastAreaZeroTime = 0
	spec.dirtyFlag = self:getNextDirtyFlag()
	
	spec.isSensorLowered = false
	spec.isManualUnloding = true
	spec.unloadState = false
	spec.restoreLastState = false
	
	self.spec_dischargeable.manualPipeDischarge = Utils.getNoNil(getXMLBool(self.xmlFile, key .. "#manualPipeDischarge"), true)
end

function Neptun:saveToXMLFile(xmlFile, key, usedModNames)
    local spec = self.spec_fruitPreparer
	setXMLBool(xmlFile, key .. "#isSensorLowered", spec.isSensorLowered)
	setXMLBool(xmlFile, key .. "#isManualUnloding", spec.isManualUnloding)
end

function Neptun:onPostLoad(savegame)
    if savegame ~= nil then
		local key = savegame.key .. "." .. Neptun.modName .. ".neptun"
		
		local isSensorLowered = Utils.getNoNil(getXMLBool(savegame.xmlFile, key .. "#isSensorLowered"), false);
		if isSensorLowered then
			self:setSensorState(isSensorLowered, true)
		end
		local isManualUnloding = Utils.getNoNil(getXMLBool(savegame.xmlFile, key .. "#isManualUnloding"), true);
		if not isManualUnloding then
			self:setUnfoldState(isManualUnloding, true)
		end
    end
end

function Neptun:loadFillEffects(xmlFile, baseKey, entry)
	entry.fillEffects = g_effectManager:loadEffect(xmlFile, baseKey .. ".fillEffect", self.components, self, self.i3dMappings)
	entry.isFilling = false
	entry.fillEnableTime = nil
	entry.fillDisableTime = nil
end
	
function Neptun:loadTerrainScanner(xmlFile, baseKey, entry)
	entry.node = I3DUtil.indexToObject(self.components, getXMLString(self.xmlFile, baseKey .. "#node"), self.i3dMappings)
    entry.width = Utils.getNoNil(getXMLFloat(xmlFile, baseKey .. "#width"), 1)
    entry.length = Utils.getNoNil(getXMLFloat(xmlFile, baseKey .. "#length"), 0.5)
    entry.yOffset = Utils.getNoNil(getXMLFloat(xmlFile, baseKey .. "#yOffset"), 0)
    entry.zOffset = Utils.getNoNil(getXMLFloat(xmlFile, baseKey .. "#zOffset"), 0)
	
	entry.animationName = Utils.getNoNil(getXMLString(xmlFile, baseKey .. "#animationName"), "")
	
    if entry.animationName ~= "" then
        local dir = -20
        self:playAnimation(entry.animationName, dir, nil, true)
        AnimatedVehicle.updateAnimations(self, 99999999)
    end	
end

function Neptun:loadDischargeParameters(xmlFile, baseKey, entry)
	local spec = self.spec_fruitPreparer
	
	entry.info = {}
	entry.info.node = I3DUtil.indexToObject(self.components, getXMLString(self.xmlFile, baseKey .. "#node"), self.i3dMappings)
    entry.info.width = Utils.getNoNil(getXMLFloat(xmlFile, baseKey .. "#width"), 1)
    entry.info.length = Utils.getNoNil(getXMLFloat(xmlFile, baseKey .. "#length"), 0.5)
    entry.info.yOffset = Utils.getNoNil(getXMLFloat(xmlFile, baseKey .. "#yOffset"), 0)
    entry.info.zOffset = Utils.getNoNil(getXMLFloat(xmlFile, baseKey .. "#zOffset"), 0)
	entry.info.limitToGround = true
	
	entry.fillUnitIndex = Utils.getNoNil(getXMLInt(xmlFile, baseKey .. "#fillUnitIndex"), spec.fillUnitIndex)
	entry.unloadInfoIndex = Utils.getNoNil(getXMLInt(xmlFile, baseKey .. "#unloadInfoIndex"), spec.unloadInfoIndex)
	
    entry.lineOffset = 0
    entry.litersToDrop = 0
	
	entry.emptySpeed = Utils.getNoNil(getXMLFloat(xmlFile, baseKey .. "#emptySpeed"), self:getFillUnitCapacity(entry.fillUnitIndex)) / 1000
	
	entry.dischargeHitTerrain = true
	
	entry.animationName = Utils.getNoNil(getXMLString(xmlFile, baseKey .. "#animationName"), "")
	entry.effects = g_effectManager:loadEffect(xmlFile, baseKey .. ".effects", self.components, self, self.i3dMappings)
	
	entry.autoDischargeMinFillLevel = Utils.getNoNil(getXMLFloat(xmlFile, baseKey .. ".autoDischarge#minFillLevel"), 0.9)
	entry.manualDischargeMinFillLevel = Utils.getNoNil(getXMLFloat(xmlFile, baseKey .. ".manualDischarge#minFillLevel"), 0.1)
	
    if self.isClient then
        entry.playSound = Utils.getNoNil(getXMLBool(xmlFile, baseKey .. "#playSound"), true)
        entry.soundNode = I3DUtil.indexToObject(self.components, getXMLString(xmlFile, baseKey .. "#soundNode"), self.i3dMappings)
        if entry.playSound then
            entry.dischargeSample = g_soundManager:loadSampleFromXML(xmlFile, baseKey, "dischargeSound", self.baseDirectory, self.components, 0, AudioGroup.VEHICLE, self.i3dMappings, self)
        end
        if Utils.getNoNil(getXMLBool(xmlFile, baseKey .. ".dischargeSound#overwriteSharedSound"), false) then
            entry.playSound = false
        end
    end
end

function Neptun:onDelete()
    if self.isClient then
        local spec = self.spec_fruitPreparer
        g_effectManager:deleteEffects(spec.fillEffects)
		g_effectManager:deleteEffects(spec.discharge.effects)
    end
end

function Neptun:onUpdateTick(dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
	local spec = self.spec_fruitPreparer
	local dischargeNode = spec.discharge
	if dischargeNode ~= nil then
		if self.isClient then
			Neptun.updateUnloadLeavesActionEvent(self)
		end
		self:updateDischargeLeavesSound(dischargeNode, dt)
		if self.isServer then
			local attacherVehicle = self:getAttacherVehicle()
			local fillLevelPercentage = self:getFillUnitFillLevelPercentage(spec.fillUnitIndex)
			if spec.isManualUnloding then
				if spec.isSensorLowered and self:getAnimationTime(spec.terrainScanner.animationName) == 1 then
					if fillLevelPercentage > spec.discharge.manualDischargeMinFillLevel then
						local sx,sy,sz = localToWorld(spec.terrainScanner.node, -spec.terrainScanner.width, spec.terrainScanner.yOffset, spec.terrainScanner.zOffset)
						local ex,ey,ez = localToWorld(spec.terrainScanner.node, spec.terrainScanner.width, spec.terrainScanner.yOffset, spec.terrainScanner.zOffset)
						local innerRadius = spec.terrainScanner.length
							
						local pickupFillType = DensityMapHeightUtil.getFillTypeAtLine(sx,sy,sz, ex,ey,ez, innerRadius)
						if pickupFillType == spec.leavesFillType then
							if not spec.unloadState then
								self:setDischargeLeavesState(true)
							end
						end
					end
				end
			else
				if fillLevelPercentage > spec.discharge.autoDischargeMinFillLevel then
					if not spec.unloadState then
						self:setDischargeLeavesState(true)
					end
				end
			end
			local allowedToDischarge = self:getCanDischargeToLand(dischargeNode) and self:getCanDischargeAtPosition(dischargeNode) and self:getCanDischargeToGround(dischargeNode) and spec.unloadState
			if allowedToDischarge then
				local fillLevel = self:getFillUnitFillLevel(dischargeNode.fillUnitIndex)
				local emptyFactor = 1
				local emptyLiters = math.min(fillLevel, dischargeNode.emptySpeed * emptyFactor * dt)
				local dischargedLiters, minDropReached, hasMinDropFillLevel = self:dischargeLeaves(dischargeNode, emptyLiters)
				spec.dischargedLiters = dischargedLiters
				local canDrop = not minDropReached and hasMinDropFillLevel
				if dischargedLiters == 0 and not canDrop then
					if spec.unloadState then
						self:setDischargeLeavesState(false)
					end
				end
			else
				if spec.unloadState then
					self:setDischargeLeavesState(false)
				end
			end
			local freeCapacity = self:getFillUnitFreeCapacity(spec.fillUnitIndex)
			spec.lastAreaZeroTime = spec.lastAreaZeroTime + dt
			if spec.isWorking and freeCapacity > 0 then
				spec.lastAreaZeroTime = 0
			end
			if spec.lastAreaZeroTime > 500 then
				if spec.fillDisableTime == nil then
					spec.fillDisableTime = g_currentMission.time
				end
			end
			if spec.fillEnableTime ~= nil and spec.fillEnableTime <= g_currentMission.time then
				self:setLeavesTankIsFilling(true, false)
				spec.fillEnableTime = nil
			end
			if spec.fillDisableTime ~= nil and spec.fillDisableTime <= g_currentMission.time then
				self:setLeavesTankIsFilling(false, false)
				spec.fillDisableTime = nil
			end
			if spec.isFilling ~= spec.sentIsFilling then
				self:raiseDirtyFlags(spec.dirtyFlag)
				spec.sentIsFilling = spec.isFilling
			end
		end
	end
end

function Neptun:dischargeLeaves(dischargeNode, emptyLiters)
    local dischargedLiters = 0
    local minDropReached = true
    local hasMinDropFillLevel = true
	dischargedLiters, minDropReached, hasMinDropFillLevel = self:dischargeToGround(dischargeNode, emptyLiters)
    return dischargedLiters, minDropReached, hasMinDropFillLevel
end

function Neptun:updateDischargeLeavesSound(dischargeNode, dt)
	local spec = self.spec_fruitPreparer
    if self.isClient then
        if spec.unloadState then
            local sharedSample = g_fillTypeManager:getSampleByFillType(spec.leavesFillType)
            if sharedSample ~= nil then
                if sharedSample ~= dischargeNode.sharedSample then
                    if dischargeNode.sample ~= nil then
                        g_soundManager:deleteSample(dischargeNode.sample)
                    end
                    dischargeNode.sample = g_soundManager:cloneSample(sharedSample, dischargeNode.info.node, self)
                    dischargeNode.sharedSample = sharedSample
                    g_soundManager:playSample(dischargeNode.sample)
                else
                    if not g_soundManager:getIsSamplePlaying(dischargeNode.sample) then
                        g_soundManager:playSample(dischargeNode.sample)
                    end
                end
            end
            if dischargeNode.dischargeSample ~= nil then
                if not g_soundManager:getIsSamplePlaying(dischargeNode.dischargeSample) then
                    g_soundManager:playSample(dischargeNode.dischargeSample)
                end
            end
            dischargeNode.turnOffSoundTimer = 500
        else
            if dischargeNode.turnOffSoundTimer ~= nil and dischargeNode.turnOffSoundTimer > 0 then
                dischargeNode.turnOffSoundTimer = dischargeNode.turnOffSoundTimer - dt
                if dischargeNode.turnOffSoundTimer <= 0 then
                    if g_soundManager:getIsSamplePlaying(dischargeNode.sample) then
                        g_soundManager:stopSample(dischargeNode.sample)
                    end
                    if dischargeNode.dischargeSample ~= nil then
                        if g_soundManager:getIsSamplePlaying(dischargeNode.dischargeSample) then
                            g_soundManager:stopSample(dischargeNode.dischargeSample)
                        end
                    end
                    dischargeNode.turnOffSoundTimer = 0
                end
            end
        end
    end
end

function Neptun:processFruitPreparerArea(superFunc, workArea)
    local spec = self.spec_fruitPreparer
    local workAreaSpec = self.spec_workArea
	
    local xs,_,zs = getWorldTranslation(workArea.start)
    local xw,_,zw = getWorldTranslation(workArea.width)
    local xh,_,zh = getWorldTranslation(workArea.height)
    local dxs, dzs = xs, zs
    local dxw, dzw = xw, zw
    local dxh, dzh = xh, zh
	
    if workArea.dropWorkAreaIndex ~= nil then
        local dropArea = workAreaSpec.workAreas[workArea.dropWorkAreaIndex]
        if dropArea ~= nil then
            dxs, _, dzs = getWorldTranslation(dropArea.start)
            dxw, _, dzw = getWorldTranslation(dropArea.width)
            dxh, _, dzh = getWorldTranslation(dropArea.height)
        end
    end
	
	local freeCapacity = self:getFillUnitFreeCapacity(spec.fillUnitIndex)
	if freeCapacity > 0 then
		local i = 0
		dxs,dzs, dxw,dzw, dxh,dzh = i,i,i,i,i,i
	end

	local area = FSDensityMapUtil.updateFruitPreparerArea(spec.fruitType, xs,zs, xw,zw, xh,zh, dxs,dzs, dxw,dzw, dxh,dzh)
	if area > 0 then
		spec.isWorking = true
		
		local pixelToSqm = g_currentMission:getFruitPixelsToSqm()
		local dropLiters = area * pixelToSqm * spec.fillMultiplier
		local loadInfo = self:getFillVolumeLoadInfo(spec.loadInfoIndex)
		self:addFillUnitFillLevel(self:getOwnerFarmId(), spec.fillUnitIndex, dropLiters, spec.leavesFillType, ToolType.UNDEFINED, loadInfo)
		
        if spec.fillEnableTime == nil then
            spec.fillEnableTime = g_currentMission.time
        end
    end
    return 0, area
end

function Neptun:setLeavesTankIsFilling(isFilling, isSynchronized)
	local spec = self.spec_fruitPreparer
	if spec.isFilling ~= isFilling then
        spec.isFilling = isFilling
        if self.isServer and isSynchronized then
            spec.sentIsFilling = isFilling
        end
        if spec.fillEffects ~= nil and not isFilling then
            g_effectManager:stopEffects(spec.fillEffects)
        end
        if isFilling then
            if spec.fillEffects ~= nil then
                g_effectManager:setFillType(spec.fillEffects, spec.leavesFillType)
                g_effectManager:startEffects(spec.fillEffects)
            end
        end
    end
end

function Neptun:getCanSensorPickupState()
	local spec = self.spec_fruitPreparer
    return spec.isManualUnloding
end

function Neptun:setSensorState(isSensorLowered, noEventSend)
    local spec = self.spec_fruitPreparer
    if isSensorLowered ~= spec.isSensorLowered then
        SetSensorStateEvent.sendEvent(self, isSensorLowered, noEventSend)
        spec.isSensorLowered = isSensorLowered
		local animName = spec.terrainScanner.animationName
        if animName ~= "" then
            local animTime = nil
            if self:getIsAnimationPlaying(animName) then
                animTime = self:getAnimationTime(animName)
            end
            if isSensorLowered then
                self:playAnimation(animName, 1, animTime, true)
            else
                self:playAnimation(animName, -1, animTime, true)
            end
        end
        Neptun.updateFoldStateActionEvent(self)
    end
end

function Neptun:setUnfoldState(isManualUnloding, noEventSend)
	local spec = self.spec_fruitPreparer
	if isManualUnloding ~= spec.isManualUnloding then
		SetUnfoldStateEvent.sendEvent(self, isManualUnloding, noEventSend)
		spec.isManualUnloding = isManualUnloding
		if spec.isSensorLowered then
			self:setSensorState(false)
		end
		if not isManualUnloding then
			if spec.unloadState then
				self:setDischargeLeavesState(false)
			end
		end
		Neptun.updateUnloadModeActionEvent(self)
		Neptun.updateFoldStateActionEvent(self)
	end
end

function Neptun:setDischargeLeavesState(state, noEventSend)
	local spec = self.spec_fruitPreparer
	local dischargeNode = spec.discharge	
	if state ~= spec.unloadState then
		SetUnloadStateEvent.sendEvent(self, state, noEventSend)
		spec.unloadState = state
		if dischargeNode.animationName ~= "" then
            local animTime = nil
            if self:getIsAnimationPlaying(spec.animationName) then
                animTime = self:getAnimationTime(spec.animationName)
            end
			if state then
				self:playAnimation(dischargeNode.animationName, 1, animTime, true)
			else
				self:playAnimation(dischargeNode.animationName, -1, animTime, true)
			end
		end
		if dischargeNode.effects ~= nil then
			if state then
				g_effectManager:setFillType(dischargeNode.effects, spec.leavesFillType)
				g_effectManager:startEffects(dischargeNode.effects)
			else
				g_effectManager:stopEffects(dischargeNode.effects)
			end
		end
	end
end

function Neptun:onTurnedOn()
end

function Neptun:onTurnedOff()
	 local spec = self.spec_fruitPreparer
	self:setLeavesTankIsFilling(false, true)
    spec.fillEnableTime = nil
    spec.fillDisableTime = nil
	if spec.unloadState then
		self:setDischargeLeavesState(false)
	end
end

function Neptun:onReadStream(streamId, connection)
    local spec = self.spec_fruitPreparer
    local tankIsFilling = streamReadBool(streamId)
	local isSensorLowered = streamReadBool(streamId)
	local isManualUnloding = streamReadBool(streamId)
	local unloadState = streamReadBool(streamId)
    self:setLeavesTankIsFilling(tankIsFilling, true)
	self:setSensorState(isSensorLowered, true)
	self:setUnfoldState(isManualUnloding, true)
	self:setDischargeLeavesState(unloadState, true)
end

function Neptun:onWriteStream(streamId, connection)
    local spec = self.spec_fruitPreparer
    streamWriteBool(streamId, spec.isFilling)
	streamWriteBool(streamId, spec.isSensorLowered)
	streamWriteBool(streamId, spec.isManualUnloding)
	streamWriteBool(streamId, spec.unloadState)
end

function Neptun:onReadUpdateStream(streamId, timestamp, connection)
    if connection:getIsServer() then
        if streamReadBool(streamId) then
            local spec = self.spec_fruitPreparer
            local tankIsFilling = streamReadBool(streamId)
            self:setLeavesTankIsFilling(tankIsFilling, true)
        end
    end
end

function Neptun:onWriteUpdateStream(streamId, connection, dirtyMask)
    if not connection:getIsServer() then
        local spec = self.spec_fruitPreparer
        if streamWriteBool(streamId, bitAND(dirtyMask, spec.dirtyFlag) ~= 0) then
            streamWriteBool(streamId, spec.isFilling)
        end
    end
end

function Neptun:getCanToggleDischargeToObject(superFunc)
	local spec = self.spec_dischargeable
	local retValue = superFunc(self)
	if spec.manualPipeDischarge then
		if not self:getIsAIActive() then
			retValue = true
		end
	end
	return retValue
end

function Neptun:handleDischargeRaycast(superFunc, dischargeNode, object, shape, distance, fillUnitIndex, hitTerrain)
	local spec = self.spec_dischargeable
	local manualDischarge = false
	if spec.manualPipeDischarge then
		if not self:getIsAIActive() then
			manualDischarge = true
		end
	end
	if manualDischarge then
		if object == nil and spec.currentDischargeState == Dischargeable.DISCHARGE_STATE_OBJECT then
			self:setDischargeState(Dischargeable.DISCHARGE_STATE_OFF)
		end
	else
		superFunc(self, dischargeNode, object, shape, distance, fillUnitIndex, hitTerrain)
	end
end